@@ -1,11 +1,14 @@
+// SPDX-License-Identifier: Apache-2.0
+// Copyright Contributors to the OpenTimelineIO project
+
 #pragma once
 
-#include <any>
+#include "any/any.hpp"
 #include "opentimelineio/version.h"
 
 namespace opentimelineio { namespace OPENTIMELINEIO_VERSION {
 
-using std::any;
-using std::any_cast;
+using linb::any;
+using linb::any_cast;
 
 }} // namespace opentimelineio::OPENTIMELINEIO_VERSION
