@@ -1,37 +1,24 @@
-#
+# SPDX-License-Identifier: Apache-2.0
 # Copyright Contributors to the OpenTimelineIO project
-#
-# Licensed under the Apache License, Version 2.0 (the "Apache License")
-# with the following modification; you may not use this file except in
-# compliance with the Apache License and the following modification to it:
-# Section 6. Trademarks. is deleted and replaced with:
-#
-# 6. Trademarks. This License does not grant permission to use the trade
-#    names, trademarks, service marks, or product names of the Licensor
-#    and its affiliates, except as required to comply with Section 4(c) of
-#    the License and to reproduce the content of the NOTICE file.
-#
-# You may obtain a copy of the Apache License at
-#
-#     http://www.apache.org/licenses/LICENSE-2.0
-#
-# Unless required by applicable law or agreed to in writing, software
-# distributed under the Apache License with the above modification is
-# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
-# KIND, either express or implied. See the Apache License for the specific
-# language governing permissions and limitations under the Apache License.
-#
 
 """OpenTimelineIO Advanced Authoring Format (AAF) Adapter
 
 Depending on if/where PyAAF is installed, you may need to set this env var:
     OTIO_AAF_PYTHON_LIB - should point at the PyAAF module.
 """
-
+import colorsys
+import copy
+import numbers
 import os
 import sys
-import numbers
-import copy
+
+try:
+    # Python 2
+    text_type = unicode
+except NameError:
+    # Python 3
+    text_type = str
+
 try:
     # Python 3.3+
     import collections.abc as collections_abc
@@ -49,6 +36,7 @@ import aaf2.content  # noqa: E402
 import aaf2.mobs  # noqa: E402
 import aaf2.components  # noqa: E402
 import aaf2.core  # noqa: E402
+import aaf2.misc  # noqa: E402
 from opentimelineio_contrib.adapters.aaf_adapter import aaf_writer  # noqa: E402
 
 
@@ -57,6 +45,16 @@ debug = False
 # If enabled, output recursive traversal info of _transcribe() method.
 _TRANSCRIBE_DEBUG = False
 
+# bake keyframed parameter
+_BAKE_KEYFRAMED_PROPERTIES_VALUES = False
+
+_PROPERTY_INTERPOLATION_MAP = {
+    aaf2.misc.ConstantInterp: "Constant",
+    aaf2.misc.LinearInterp: "Linear",
+    aaf2.misc.BezierInterpolator: "Bezier",
+    aaf2.misc.CubicInterpolator: "Cubic",
+}
+
 
 def _transcribe_log(s, indent=0, always_print=False):
     if always_print or _TRANSCRIBE_DEBUG:
@@ -100,34 +98,185 @@ def _get_class_name(item):
         return item.__class__.__name__
 
 
-def _transcribe_property(prop):
+def _transcribe_property(prop, owner=None):
     # XXX: The unicode type doesn't exist in Python 3 (all strings are unicode)
     # so we have to use type(u"") which works in both Python 2 and 3.
-    if isinstance(prop, (str, type(u""), numbers.Integral, float)):
+    if isinstance(prop, (str, type(u""), numbers.Integral, float, dict)):
         return prop
-
+    elif isinstance(prop, set):
+        return list(prop)
     elif isinstance(prop, list):
         result = {}
         for child in prop:
-            if hasattr(child, "name") and hasattr(child, "value"):
-                result[child.name] = _transcribe_property(child.value)
+            if hasattr(child, "name"):
+                if isinstance(child, aaf2.misc.VaryingValue):
+                    # keyframed values
+                    control_points = []
+                    for control_point in child["PointList"]:
+                        try:
+                            # Some values cannot be transcribed yet
+                            control_points.append(
+                                [
+                                    control_point.time,
+                                    _transcribe_property(control_point.value),
+                                ]
+                            )
+                        except TypeError:
+                            _transcribe_log(
+                                "Unable to transcribe value for property: "
+                                "'{}' (Type: '{}', Parent: '{}')".format(
+                                    child.name, type(child), prop
+                                )
+                            )
+
+                    # bake keyframe values for owner time range
+                    baked_values = None
+                    if _BAKE_KEYFRAMED_PROPERTIES_VALUES:
+                        if isinstance(owner, aaf2.components.Component):
+                            baked_values = []
+                            for t in range(0, owner.length):
+                                baked_values.append([t, child.value_at(t)])
+                        else:
+                            _transcribe_log(
+                                "Unable to bake values for property: "
+                                "'{}'. Owner: {}, Control Points: {}".format(
+                                    child.name, owner, control_points
+                                )
+                            )
+
+                    value_dict = {
+                        "_aaf_keyframed_property": True,
+                        "keyframe_values": control_points,
+                        "keyframe_interpolation": _PROPERTY_INTERPOLATION_MAP.get(
+                            child.interpolationdef.auid, "Linear"
+                        ),
+                        "keyframe_baked_values": baked_values
+                    }
+                    result[child.name] = value_dict
+
+                elif hasattr(child, "value"):
+                    # static value
+                    result[child.name] = _transcribe_property(child.value, owner=owner)
             else:
                 # @TODO: There may be more properties that we might want also.
                 # If you want to see what is being skipped, turn on debug.
                 if debug:
-                    debug_message = \
-                        "Skipping unrecognized property: {} of parent {}"
-                    print(debug_message.format(child, prop))
+                    debug_message = "Skipping unrecognized property: '{}', parent '{}'"
+                    _transcribe_log(debug_message.format(child, prop))
         return result
     elif hasattr(prop, "properties"):
         result = {}
         for child in prop.properties():
-            result[child.name] = _transcribe_property(child.value)
+            result[child.name] = _transcribe_property(child.value, owner=owner)
         return result
     else:
         return str(prop)
 
 
+def _otio_color_from_hue(hue):
+    """Return an OTIO marker color, based on hue in range of [0.0, 1.0].
+
+    Args:
+        hue (float): marker color hue value
+
+    Returns:
+        otio.schema.MarkerColor: converted / estimated marker color
+
+    """
+    if hue <= 0.04 or hue > 0.93:
+        return otio.schema.MarkerColor.RED
+    if hue <= 0.13:
+        return otio.schema.MarkerColor.ORANGE
+    if hue <= 0.2:
+        return otio.schema.MarkerColor.YELLOW
+    if hue <= 0.43:
+        return otio.schema.MarkerColor.GREEN
+    if hue <= 0.52:
+        return otio.schema.MarkerColor.CYAN
+    if hue <= 0.74:
+        return otio.schema.MarkerColor.BLUE
+    if hue <= 0.82:
+        return otio.schema.MarkerColor.PURPLE
+    return otio.schema.MarkerColor.MAGENTA
+
+
+def _marker_color_from_string(color):
+    """Tries to derive a valid marker color from a string.
+
+    Args:
+        color (str): color name (e.g. "Yellow")
+
+    Returns:
+        otio.schema.MarkerColor: matching color or `None`
+    """
+    if not color:
+        return
+
+    return getattr(otio.schema.MarkerColor, color.upper(), None)
+
+
+def _convert_rgb_to_marker_color(rgb_dict):
+    """Returns a matching OTIO marker color for a given AAF color string.
+
+    Adapted from `get_nearest_otio_color()` in the `xges.py` adapter.
+
+    Args:
+        rgb_dict (dict): marker color as dict,
+                         e.g. `"{'red': 41471, 'green': 12134, 'blue': 6564}"`
+
+    Returns:
+        otio.schema.MarkerColor: converted / estimated marker color
+
+    """
+
+    float_colors = {
+        (1.0, 0.0, 0.0): otio.schema.MarkerColor.RED,
+        (0.0, 1.0, 0.0): otio.schema.MarkerColor.GREEN,
+        (0.0, 0.0, 1.0): otio.schema.MarkerColor.BLUE,
+        (0.0, 0.0, 0.0): otio.schema.MarkerColor.BLACK,
+        (1.0, 1.0, 1.0): otio.schema.MarkerColor.WHITE,
+    }
+
+    # convert from UInt to float
+    red = float(rgb_dict["red"]) / 65535.0
+    green = float(rgb_dict["green"]) / 65535.0
+    blue = float(rgb_dict["blue"]) / 65535.0
+    rgb_float = (red, green, blue)
+
+    # check for exact match
+    marker_color = float_colors.get(rgb_float)
+    if marker_color:
+        return marker_color
+
+    # try to get an approxiate match based on hue
+    hue, lightness, saturation = colorsys.rgb_to_hls(red, green, blue)
+    nearest = None
+    if saturation < 0.2:
+        if lightness > 0.65:
+            nearest = otio.schema.MarkerColor.WHITE
+        else:
+            nearest = otio.schema.MarkerColor.BLACK
+    if nearest is None:
+        if lightness < 0.13:
+            nearest = otio.schema.MarkerColor.BLACK
+        if lightness > 0.9:
+            nearest = otio.schema.MarkerColor.WHITE
+    if nearest is None:
+        nearest = _otio_color_from_hue(hue)
+        if nearest == otio.schema.MarkerColor.RED and lightness > 0.53:
+            nearest = otio.schema.MarkerColor.PINK
+        if (
+            nearest == otio.schema.MarkerColor.MAGENTA
+            and hue < 0.89
+            and lightness < 0.42
+        ):
+            # some darker magentas look more like purple
+            nearest = otio.schema.MarkerColor.PURPLE
+
+    # default to red color
+    return nearest or otio.schema.MarkerColor.RED
+
+
 def _find_timecode_mobs(item):
     mobs = [item.mob]
 
@@ -252,7 +401,7 @@ def _transcribe(item, parents, edit_rate, indent=0):
             if hasattr(prop, 'name') and hasattr(prop, 'value'):
                 key = str(prop.name)
                 value = prop.value
-                metadata[key] = _transcribe_property(value)
+                metadata[key] = _transcribe_property(value, owner=item)
 
     # Now we will use the item's class to determine which OTIO type
     # to transcribe into. Note that the order of this if/elif/... chain
@@ -474,6 +623,17 @@ def _transcribe(item, parents, edit_rate, indent=0):
         _transcribe_log(msg, indent)
         result = otio.schema.Track()
 
+        # if parent is a sequence add SlotID / PhysicalTrackNumber to attach markers
+        parent = parents[-1]
+        if isinstance(parent, (aaf2.components.Sequence, aaf2.components.NestedScope)):
+            timeline_slots = [
+                p for p in parents if isinstance(p, aaf2.mobslots.TimelineMobSlot)
+            ]
+            timeline_slot = timeline_slots[-1]
+            if timeline_slot:
+                metadata["PhysicalTrackNumber"] = list(parent.slots).index(item) + 1
+                metadata["SlotID"] = int(timeline_slot["SlotID"].value)
+
         for component in item.components:
             child = _transcribe(component, parents + [item], edit_rate, indent + 2)
             _add_child(result, child, component)
@@ -524,32 +684,100 @@ def _transcribe(item, parents, edit_rate, indent=0):
         )
 
     elif isinstance(item, aaf2.components.DescriptiveMarker):
+        event_mobs = [p for p in parents if isinstance(p, aaf2.mobslots.EventMobSlot)]
+        if event_mobs:
+            _transcribe_log(
+                "Create marker for '{}'".format(_encoded_name(item)), indent
+            )
 
-        # Markers come in on their own separate Track.
-        # TODO: We should consolidate them onto the same track(s) as the clips
-        # result = otio.schema.Marker()
-        pass
+            result = otio.schema.Marker()
+            result.name = metadata["Comment"]
+
+            event_mob = event_mobs[-1]
+
+            metadata["AttachedSlotID"] = int(metadata["DescribedSlots"][0])
+            metadata["AttachedPhysicalTrackNumber"] = int(
+                event_mob["PhysicalTrackNumber"].value
+            )
+
+            # determine marker color
+            color = _marker_color_from_string(
+                metadata.get("CommentMarkerAttributeList", {}).get("_ATN_CRM_COLOR")
+            )
+            if color is None:
+                color = _convert_rgb_to_marker_color(
+                    metadata["CommentMarkerColor"]
+                )
+            result.color = color
+
+            position = metadata["Position"]
+
+            # Length can be None, but the property will always exist
+            # so get('Length', 1) wouldn't help.
+            length = metadata["Length"]
+            if length is None:
+                length = 1
+
+            result.marked_range = otio.opentime.TimeRange(
+                start_time=otio.opentime.from_frames(position, edit_rate),
+                duration=otio.opentime.from_frames(length, edit_rate),
+            )
+        else:
+            _transcribe_log(
+                "Cannot attach marker item '{}'. "
+                "Missing event mob in hierarchy.".format(
+                    _encoded_name(item)
+                )
+            )
 
     elif isinstance(item, aaf2.components.Selector):
         msg = "Transcribe selector for  {}".format(_encoded_name(item))
         _transcribe_log(msg, indent)
-        # If you mute a clip in media composer, it becomes one of these in the
-        # AAF.
-        result = _transcribe(item.getvalue("Selected"),
-                             parents + [item], edit_rate, indent + 2)
 
-        alternates = [
-            _transcribe(alt, parents + [item], edit_rate, indent + 2)
-            for alt in item.getvalue("Alternates")
-        ]
+        selected = item.getvalue('Selected')
+        alternates = item.getvalue('Alternates', None)
 
-        # muted case -- if there is only one item its muted, otherwise its
-        # a multi cam thing
-        if alternates and len(alternates) == 1:
-            metadata['muted_clip'] = True
-            result.name = str(alternates[0].name) + "_MUTED"
+        # First we check to see if the Selected component is either a Filler
+        # or ScopeReference object, meaning we have to use the alternate instead
+        if isinstance(selected, aaf2.components.Filler) or \
+                isinstance(selected, aaf2.components.ScopeReference):
 
-        metadata['alternates'] = alternates
+            # Safety check of the alternates list, then transcribe first object -
+            # there should only ever be one alternate in this situation
+            if alternates is None or len(alternates) != 1:
+                err = "AAF Selector parsing error: object has unexpected number of " \
+                      "alternates - {}".format(len(alternates))
+                raise AAFAdapterError(err)
+            result = _transcribe(alternates[0], parents + [item], edit_rate, indent + 2)
+
+            # Filler/ScopeReference means the clip is muted/not enabled
+            result.enabled = False
+
+            # Muted tracks are handled in a slightly odd way so we need to do a
+            # check here and pass the param back up to the track object
+            # if isinstance(parents[-1], aaf2.mobslots.TimelineMobSlot):
+            #     pass # TODO: Figure out mechanism for passing this up to parent
+
+        else:
+
+            # This is most likely a multi-cam clip
+            result = _transcribe(selected, parents + [item], edit_rate, indent + 2)
+
+            # Perform a check here to make sure no potential Gap objects
+            # are slipping through the cracks
+            if isinstance(result, otio.schema.Gap):
+                err = "AAF Selector parsing error: {}".format(type(item))
+                raise AAFAdapterError(err)
+
+            # A Selector can have a set of alternates to handle multiple options for an
+            # editorial decision - we do a full parse on those obects too
+            if alternates is not None:
+                alternates = [
+                    _transcribe(alt, parents + [item], edit_rate, indent + 2)
+                    for alt in alternates
+                ]
+
+            metadata['alternates'] = alternates
 
     # @TODO: There are a bunch of other AAF object types that we will
     # likely need to add support for. I'm leaving this code here to help
@@ -986,6 +1214,103 @@ def _fix_transitions(thing):
             _fix_transitions(child)
 
 
+def _attach_markers(collection):
+    """Search for markers on tracks and attach them to their corresponding item.
+
+    Marked ranges will also be transformed into the new parent space.
+
+    """
+    # iterate all timeline objects
+    for timeline in collection.each_child(descended_from_type=otio.schema.Timeline):
+        tracks_map = {}
+
+        # build track mapping
+        for track in timeline.each_child(descended_from_type=otio.schema.Track):
+            metadata = track.metadata.get("AAF", {})
+            slot_id = metadata.get("SlotID")
+            track_number = metadata.get("PhysicalTrackNumber")
+            if slot_id is None or track_number is None:
+                continue
+
+            tracks_map[(int(slot_id), int(track_number))] = track
+
+        # iterate all tracks for their markers and attach them to the matching item
+        for current_track in timeline.each_child(descended_from_type=otio.schema.Track):
+            for marker in list(current_track.markers):
+                metadata = marker.metadata.get("AAF", {})
+                slot_id = metadata.get("AttachedSlotID")
+                track_number = metadata.get("AttachedPhysicalTrackNumber")
+                target_track = tracks_map.get((slot_id, track_number))
+                if target_track is None:
+                    raise AAFAdapterError(
+                        "Marker '{}' cannot be attached to an item. SlotID: '{}', "
+                        "PhysicalTrackNumber: '{}'".format(
+                            marker.name, slot_id, track_number
+                        )
+                    )
+
+                # remove marker from current parent track
+                current_track.markers.remove(marker)
+
+                # determine new item to attach the marker to
+                try:
+                    target_item = target_track.child_at_time(
+                        marker.marked_range.start_time
+                    )
+
+                    if target_item is None or not hasattr(target_item, 'markers'):
+                        # Item found cannot have markers, for example Transition.
+                        # See also `marker-over-transition.aaf` in test data.
+                        #
+                        # Leave markers on the track for now.
+                        _transcribe_log(
+                            'Skip target_item `{}` cannot have markers'.format(
+                                target_item,
+                            ),
+                        )
+                        target_item = target_track
+
+                    # transform marked range into new item range
+                    marked_start_local = current_track.transformed_time(
+                        marker.marked_range.start_time, target_item
+                    )
+
+                    marker.marked_range = otio.opentime.TimeRange(
+                        start_time=marked_start_local,
+                        duration=marker.marked_range.duration
+                    )
+
+                except otio.exceptions.CannotComputeAvailableRangeError as e:
+                    # For audio media AAF file (marker-over-audio.aaf),
+                    # this exception would be triggered in:
+                    # `target_item = target_track.child_at_time()` with error
+                    # message:
+                    # "No available_range set on media reference on clip".
+                    #
+                    # Leave markers on the track for now.
+                    _transcribe_log(
+                        'Cannot compute availableRange from {} to {}: {}'.format(
+                            marker,
+                            target_track,
+                            e,
+                        ),
+                    )
+                    target_item = target_track
+
+                # attach marker to target item
+                target_item.markers.append(marker)
+
+                _transcribe_log(
+                    "Marker: '{}' (time: {}), attached to item: '{}'".format(
+                        marker.name,
+                        marker.marked_range.start_time.value,
+                        target_item.name,
+                    )
+                )
+
+    return collection
+
+
 def _simplify(thing):
     # If the passed in is an empty dictionary or None, nothing to do.
     # Without this check it would still return thing, but this way we avoid
@@ -1060,6 +1385,9 @@ def _simplify(thing):
                     # Note: we don't merge effects, because we already made
                     # sure the child had no effects in the if statement above.
 
+                    # Preserve the enabled/disabled state as we merge these two.
+                    thing.enabled = thing.enabled and child.enabled
+
                     c = c + num
                 c = c - 1
 
@@ -1067,8 +1395,16 @@ def _simplify(thing):
         if _is_redundant_container(thing):
             # TODO: We may be discarding metadata here, should we merge it?
             result = thing[0].deepcopy()
+
+            # As we are reducing the complexity of the object structure through
+            # this process, we need to make sure that any/all enabled statuses
+            # are being respected and applied in an appropriate way
+            if not thing.enabled:
+                result.enabled = False
+
             # TODO: Do we need to offset the markers in time?
             result.markers.extend(thing.markers)
+
             # TODO: The order of the effects is probably important...
             # should they be added to the end or the front?
             # Intuitively it seems like the child's effects should come before
@@ -1167,32 +1503,94 @@ def _contains_something_valuable(thing):
     return True
 
 
-def read_from_file(filepath, simplify=True, transcribe_log=False):
+def _get_mobs_for_transcription(storage):
+    """
+    When we describe our AAF into OTIO space, we apply the following heuristic:
 
+    1) First look for top level mobs and if found use that to transcribe.
+
+    2) If we don't have top level mobs, look for composition mobs and use them to
+    transcribe.
+
+    3) Lastly if we don't have either, try to use master mobs to transcribe.
+
+    If we don't find any Mobs, just tell the user and do transcrption on an empty
+    list (to generate some 'empty-level' OTIO structure)
+
+    This heuristic is based on 'real-world' examples. There may still be some
+    corner cases / open questions (like could there be metadata on both
+    a composition mob and master mob? And if so, who would 'win'?)
+
+    In any way, this heuristic satisfies the current set of AAFs we are using
+    in our test-environment.
+
+    """
+
+    top_level_mobs = list(storage.toplevel())
+
+    if len(top_level_mobs) > 0:
+        _transcribe_log("---\nTranscribing top level mobs\n---")
+        return top_level_mobs
+
+    composition_mobs = list(storage.compositionmobs())
+    if len(composition_mobs) > 0:
+        _transcribe_log("---\nTranscribing composition mobs\n---")
+        return composition_mobs
+
+    master_mobs = list(storage.mastermobs())
+    if len(master_mobs) > 0:
+        _transcribe_log("---\nTranscribing master mobs\n---")
+        return master_mobs
+
+    _transcribe_log("---\nNo mobs found to transcribe\n---")
+
+    return []
+
+
+def read_from_file(
+    filepath,
+    simplify=True,
+    transcribe_log=False,
+    attach_markers=True,
+    bake_keyframed_properties=False
+):
+    """Reads AAF content from `filepath` and outputs an OTIO timeline object.
+
+    Args:
+        filepath (str): AAF filepath
+        simplify (bool, optional): simplify timeline structure by stripping empty items
+        transcribe_log (bool, optional): log activity as items are getting transcribed
+        attach_markers (bool, optional): attaches markers to their appropriate items
+                                         like clip, gap. etc on the track
+        bake_keyframed_properties (bool, optional): bakes animated property values
+                                                    for each frame in a source clip
+    Returns:
+        otio.schema.Timeline
+
+    """
     # 'activate' transcribe logging if adapter argument is provided.
     # Note that a global 'switch' is used in order to avoid
     # passing another argument around in the _transcribe() method.
     #
-    global _TRANSCRIBE_DEBUG
+    global _TRANSCRIBE_DEBUG, _BAKE_KEYFRAMED_PROPERTIES_VALUES
     _TRANSCRIBE_DEBUG = transcribe_log
+    _BAKE_KEYFRAMED_PROPERTIES_VALUES = bake_keyframed_properties
 
     with aaf2.open(filepath) as aaf_file:
+        # Note: We're skipping: aaf_file.header
+        # Is there something valuable in there?
 
         storage = aaf_file.content
+        mobs_to_transcribe = _get_mobs_for_transcription(storage)
 
-        # Note: We're skipping: f.header
-        # Is there something valuable in there?
-        _transcribe_log("---\nTranscribing top level mobs\n---", 0)
+        result = _transcribe(mobs_to_transcribe, parents=list(), edit_rate=None)
 
-        # Get just the top-level MOBS from the AAF
-        top = list(storage.toplevel())
-
-        # Transcribe just the top-level mobs
-        result = _transcribe(top, parents=list(), edit_rate=None)
+    # Attach marker to the appropriate clip, gap etc.
+    if attach_markers:
+        result = _attach_markers(result)
 
     # AAF is typically more deeply nested than OTIO.
-
-    # Lets try to simplify the structure by collapsing or removing
+    # Let's try to simplify the structure by collapsing or removing
     # unnecessary stuff.
     if simplify:
         result = _simplify(result)
